import gc
import warnings
from typing import Union, Callable
import logging
import os
import time

import datetime
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy import signal
import xarray as xr
import dask.array as da

from mlair.helpers import to_list, TimeTrackingWrapper, TimeTracking


class FIRFilter:

    def __init__(self, data, fs, order, cutoff, window, dim):

        filtered = []
        h = []
        for i in range(len(order)):
            fi, hi = fir_filter(data, fs, order=order[i], cutoff_low=cutoff[i][0], cutoff_high=cutoff[i][1],
                                window=window, dim=dim, h=None, causal=True, padlen=None)
            filtered.append(fi)
            h.append(hi)

        self._filtered = filtered
        self._h = h

    def filter_coefficients(self):
        return self._h

    def filtered_data(self):
        return self._filtered
        #
        # y, h = fir_filter(station_data.values.flatten(), fs, order[0], cutoff_low=cutoff[0][0], cutoff_high=cutoff[0][1],
        #                   window=window)
        # filtered = xr.ones_like(station_data) * y.reshape(station_data.values.shape)
        # # band pass
        # y_band, h_band = fir_filter(station_data.values.flatten(), fs, order[1], cutoff_low=cutoff[1][0],
        #                             cutoff_high=cutoff[1][1], window=window)
        # filtered_band = xr.ones_like(station_data) * y_band.reshape(station_data.values.shape)
        # # band pass 2
        # y_band_2, h_band_2 = fir_filter(station_data.values.flatten(), fs, order[2], cutoff_low=cutoff[2][0],
        #                                 cutoff_high=cutoff[2][1], window=window)
        # filtered_band_2 = xr.ones_like(station_data) * y_band_2.reshape(station_data.values.shape)
        # # high pass
        # y_high, h_high = fir_filter(station_data.values.flatten(), fs, order[3], cutoff_low=cutoff[3][0],
        #                             cutoff_high=cutoff[3][1], window=window)
        # filtered_high = xr.ones_like(station_data) * y_high.reshape(station_data.values.shape)


class ClimateFIRFilter:

    def __init__(self, data, fs, order, cutoff, window, time_dim, var_dim, apriori=None, apriori_type=None,
                 apriori_diurnal=False, sel_opts=None, plot_path=None, plot_name=None, vectorized=True,
                 padlen_factor=0.8, minimum_length=None, new_dim=None):
        """
        :param data: data to filter
        :param fs: sampling frequency in 1/days -> 1d: fs=1 -> 1H: fs=24
        :param order: a tuple with the order of the filter in same ordering like cutoff
        :param cutoff: a tuple with the cutoff frequencies (all are applied as low pass)
        :param window: window type of the filter (e.g. hamming)
        :param time_dim: name of time dimension to apply filter along
        :param var_dim: name of variables dimension
        :param apriori: apriori information to use for the first low pass. If None, climatology is calculated on the
            provided data.
        :param apriori_type: type of apriori information to use. Climatology will be used always for first low pass. For
            the residuum either the value zero is used (apriori_type is None or "zeros") or a climatology on the
            residua is used ("residuum_stats").
        :param apriori_diurnal: Use diurnal cycle as additional apriori information (only applicable for hourly
            resoluted data). The mean anomaly of each hour is added to the apriori_type information.
        """
        logging.info(f"{plot_name}: start init ClimateFIRFilter")
        self.plot_path = plot_path
        self.plot_name = plot_name
        self.plot_data = []
        filtered = []
        h = []
        sel_opts = sel_opts if isinstance(sel_opts, dict) else {time_dim: sel_opts}
        sampling = {1: "1d", 24: "1H"}.get(int(fs))
        logging.info(f"{plot_name}: create diurnal_anomalies")
        if apriori_diurnal is True and sampling == "1H":
            diurnal_anomalies = self.create_hourly_mean(data, sel_opts=sel_opts, sampling=sampling, time_dim=time_dim,
                                                        as_anomaly=True)
        else:
            diurnal_anomalies = 0
        logging.info(f"{plot_name}: create monthly apriori")
        if apriori is None:
            apriori = self.create_monthly_mean(data, sel_opts=sel_opts, sampling=sampling,
                                               time_dim=time_dim) + diurnal_anomalies
            logging.info(f"{plot_name}: apriori shape = {apriori.shape}")
        apriori_list = to_list(apriori)
        input_data = data.__deepcopy__()

        # for viz
        plot_dates = None

        # create tmp dimension to apply filter, search for unused name
        new_dim = self._create_tmp_dimension(input_data) if new_dim is None else new_dim

        for i in range(len(order)):
            logging.info(f"{plot_name}: start filter for order {order[i]}")
            # calculate climatological filter
            # clim_filter: Callable = {True: self.clim_filter_vectorized, False: self.clim_filter}[vectorized]
            # ToDo: remove all methods except the vectorized version
            clim_filter: Callable = {True: self.clim_filter_vectorized_less_memory, False: self.clim_filter}[vectorized]
            _minimum_length = self._minimum_length(order, minimum_length, i)
            fi, hi, apriori, plot_data = clim_filter(input_data, fs, cutoff[i], order[i],
                                                     apriori=apriori_list[i],
                                                     sel_opts=sel_opts, sampling=sampling, time_dim=time_dim,
                                                     window=window,
                                                     var_dim=var_dim, plot_index=i, padlen_factor=padlen_factor,
                                                     minimum_length=_minimum_length, new_dim=new_dim,
                                                     plot_dates=plot_dates)

            logging.info(f"{plot_name}: finished clim_filter calculation")
            if minimum_length is None:
                filtered.append(fi)
            else:
                filtered.append(fi.sel({new_dim: slice(-minimum_length, 0)}))
            h.append(hi)
            gc.collect()
            self.plot_data.append(plot_data)
            plot_dates = {e["t0"] for e in plot_data}

            # calculate residuum
            logging.info(f"{plot_name}: calculate residuum")
            coord_range = range(fi.coords[new_dim].values.min(), fi.coords[new_dim].values.max() + 1)
            if new_dim in input_data.coords:
                input_data = input_data.sel({new_dim: coord_range}) - fi
            else:
                input_data = self._shift_data(input_data, coord_range, time_dim, var_dim, new_dim) - fi

            # create new apriori information for next iteration if no further apriori is provided
            if len(apriori_list) <= i + 1:
                logging.info(f"{plot_name}: create diurnal_anomalies")
                if apriori_diurnal is True and sampling == "1H":
                    diurnal_anomalies = self.create_hourly_mean(input_data.sel({new_dim: 0}, drop=True),
                                                                sel_opts=sel_opts, sampling=sampling,
                                                                time_dim=time_dim, as_anomaly=True)
                else:
                    diurnal_anomalies = 0
                logging.info(f"{plot_name}: create monthly apriori")
                if apriori_type is None or apriori_type == "zeros":  # zero version
                    apriori_list.append(xr.zeros_like(apriori_list[i]) + diurnal_anomalies)
                elif apriori_type == "residuum_stats":  # calculate monthly statistic on residuum
                    apriori_list.append(
                        -self.create_monthly_mean(input_data.sel({new_dim: 0}, drop=True), sel_opts=sel_opts,
                                                  sampling=sampling,
                                                  time_dim=time_dim) + diurnal_anomalies)
                else:
                    raise ValueError(f"Cannot handle unkown apriori type: {apriori_type}. Please choose from None, "
                                     f"`zeros` or `residuum_stats`.")
        # add last residuum to filtered
        if minimum_length is None:
            filtered.append(input_data)
        else:
            filtered.append(input_data.sel({new_dim: slice(-minimum_length, 0)}))
        # filtered.append(input_data)
        self._filtered = filtered
        self._h = h
        self._apriori = apriori_list

        # visualize
        self._plot(sampling)

    @staticmethod
    def _minimum_length(order, minimum_length, pos):
        next_order = 0
        if pos + 1 < len(order):
            next_order = order[pos + 1]
        if minimum_length is not None:
            next_order = next_order + minimum_length
        return next_order if next_order > 0 else None

    @staticmethod
    def create_unity_array(data, time_dim, extend_range=366):
        """Create a xr data array filled with ones. time_dim is extended by extend_range days in future and past."""
        coords = data.coords

        # extend time_dim by given extend_range days
        start = coords[time_dim][0].values.astype("datetime64[D]") - np.timedelta64(extend_range, "D")
        end = coords[time_dim][-1].values.astype("datetime64[D]") + np.timedelta64(extend_range, "D")
        new_time_axis = np.arange(start, end).astype("datetime64[ns]")

        # construct data array with updated coords
        new_coords = {k: data.coords[k].values if k != time_dim else new_time_axis for k in coords}
        new_array = xr.DataArray(1, coords=new_coords, dims=new_coords.keys()).transpose(*data.dims)

        # loffset is required because resampling uses last day in month as resampling timestamp
        return new_array.resample({time_dim: "1m"}, loffset=datetime.timedelta(days=-15)).max()

    def create_monthly_mean(self, data, sel_opts=None, sampling="1d", time_dim="datetime"):
        """Calculate monthly statistics."""

        # create unity xarray in monthly resolution with sampling point in mid of each month
        monthly = self.create_unity_array(data, time_dim)

        # apply selection if given (only use subset for monthly means)
        if sel_opts is not None:
            data = data.sel(**sel_opts)

        # create monthly mean and replace entries in unity array
        monthly_mean = data.groupby(f"{time_dim}.month").mean()
        for month in monthly_mean.month.values:
            monthly = xr.where((monthly[f"{time_dim}.month"] == month),
                               monthly_mean.sel(month=month, drop=True),
                               monthly)
        # transform monthly information into original sampling rate
        return monthly.resample({time_dim: sampling}).interpolate()

        # for month in monthly_mean.month.values:
        #     loc = (monthly[f"{time_dim}.month"] == month)
        #     monthly.loc[{time_dim: loc}] = monthly_mean.sel(month=month, drop=True)
        # aggregate monthly information (shift by half month, because resample base is last day)
        # return monthly.resample({time_dim: "1m"}).max().resample({time_dim: sampling}).interpolate()

    @staticmethod
    def create_hourly_mean(data, sel_opts=None, sampling="1H", time_dim="datetime", as_anomaly=True):
        """Calculate hourly statistics. Either the absolute value or the anomaly (as_anomaly=True)."""
        # can only be used for hourly sampling rate
        assert sampling == "1H"

        # create unity xarray in hourly resolution
        hourly = xr.ones_like(data)

        # apply selection if given (only use subset for hourly means)
        if sel_opts is not None:
            data = data.sel(**sel_opts)

        # create mean for each hour and replace entries in unity array, calculate anomaly if enabled
        hourly_mean = data.groupby(f"{time_dim}.hour").mean()
        if as_anomaly is True:
            hourly_mean = hourly_mean - hourly_mean.mean("hour")
        for hour in hourly_mean.hour.values:
            loc = (hourly[f"{time_dim}.hour"] == hour)
            hourly.loc[{f"{time_dim}": loc}] = hourly_mean.sel(hour=hour)
        return hourly

    @staticmethod
    def extend_apriori(data, apriori, time_dim, sampling="1d"):
        """
        Extend time range of apriori information.

        This method will fail, if apriori is available for a shorter period than the gab to fill.
        """
        dates = data.coords[time_dim].values
        td_type = {"1d": "D", "1H": "h"}.get(sampling)

        # apriori starts after data
        if dates[0] < apriori.coords[time_dim].values[0]:
            logging.info(f"{data.coords['Stations'].values[0]}: apriori starts after data")
            # add difference in full years
            date_diff = abs(dates[0] - apriori.coords[time_dim].values[0]).astype("timedelta64[D]")
            extend_range = np.ceil(date_diff / (np.timedelta64(1, "D") * 365)).astype(int) * 365
            coords = apriori.coords

            # create new time axis
            # start = coords[time_dim][0].values.astype("datetime64[%s]" % td_type) - np.timedelta64(extend_range, "D")
            # end = coords[time_dim][0].values.astype("datetime64[%s]" % td_type)
            # new_time_axis = np.arange(start, end).astype("datetime64[ns]")

            factor = 1 if td_type == "D" else 24
            start = coords[time_dim][0].values.astype("datetime64[%s]" % td_type) - np.timedelta64(
                extend_range * factor + 1,
                td_type)
            end = coords[time_dim][0].values.astype("datetime64[%s]" % td_type)
            new_time_axis = np.arange(start, end).astype("datetime64[ns]")
            logging.info(f"{data.coords['Stations'].values[0]}: shape of new_time_axis = {new_time_axis.shape}")

            # extract old values to use with new axis
            # start = coords[time_dim][0].values.astype("datetime64[D]")
            # end = coords[time_dim][0].values.astype("datetime64[D]") + np.timedelta64(extend_range - 1, "D")
            # new_values = apriori.sel({time_dim: slice(start, end)})
            # new_values.coords[time_dim] = new_time_axis

            start = coords[time_dim][0].values.astype("datetime64[%s]" % td_type)
            end = coords[time_dim][0].values.astype("datetime64[%s]" % td_type) + np.timedelta64(
                extend_range * factor - 1, td_type)
            new_values = apriori.sel({time_dim: slice(start, end)})
            logging.info(f"{data.coords['Stations'].values[0]}: shape of new_values = {new_values.shape}")
            new_values.coords[time_dim] = new_time_axis

            # add new values to apriori
            apriori = apriori.combine_first(new_values)

        # apriori ends before data
        if dates[-1] + np.timedelta64(365, "D") > apriori.coords[time_dim].values[-1]:
            logging.info(f"{data.coords['Stations'].values[0]}: apriori ends before data")
            # add difference in full years + 1 year (because apriori is used as future estimate)
            date_diff = abs(dates[-1] - apriori.coords[time_dim].values[-1]).astype("timedelta64[D]")
            extend_range = np.ceil(date_diff / (np.timedelta64(1, "D") * 365)).astype(int) * 365 + 365
            coords = apriori.coords

            # create new time axis
            factor = 1 if td_type == "D" else 24
            start = coords[time_dim][-1].values.astype("datetime64[%s]" % td_type)
            end = coords[time_dim][-1].values.astype("datetime64[%s]" % td_type) + np.timedelta64(
                extend_range * factor + 1,
                td_type)
            new_time_axis = np.arange(start, end).astype("datetime64[ns]")  # hint: arange does not include end date
            logging.info(f"{data.coords['Stations'].values[0]}: shape of new_time_axis = {new_time_axis.shape}")
            logging.info(f"{data.coords['Stations'].values[0]}: start of new_time_axis = {start}")
            logging.info(f"{data.coords['Stations'].values[0]}: end of new_time_axis = {end}")
            logging.info(f"{data.coords['Stations'].values[0]}: delta of new_time_axis = {end - start}")

            # extract old values to use with new axis
            start = coords[time_dim][-1].values.astype("datetime64[%s]" % td_type) - np.timedelta64(
                extend_range * factor, td_type)
            # start = coords[time_dim][-1].values.astype("datetime64[%s]" % td_type) - np.timedelta64(
            #     extend_range * factor, td_type)
            end = coords[time_dim][-1].values.astype("datetime64[%s]" % td_type)
            new_values = apriori.sel({time_dim: slice(start, end)})  # hint: slice includes end date
            logging.info(f"{data.coords['Stations'].values[0]}: shape of new_values = {new_values.shape}")
            logging.info(f"{data.coords['Stations'].values[0]}: start of new_values = {start}")
            logging.info(f"{data.coords['Stations'].values[0]}: end of new_values = {end}")
            logging.info(f"{data.coords['Stations'].values[0]}: delta of new_values = {end - start}")

            logging.info(f"{data.coords['Stations'].values[0]}: set new_time_axis")
            new_values.coords[time_dim] = new_time_axis

            # add new values to apriori
            logging.info(f"{data.coords['Stations'].values[0]}: add to apriori")
            apriori = apriori.combine_first(new_values)

        return apriori

    @TimeTrackingWrapper
    def clim_filter(self, data, fs, cutoff_high, order, apriori=None, padlen_factor=0.5, sel_opts=None, sampling="1d",
                    time_dim="datetime", var_dim="variables", window="hamming", plot_index=None):

        # calculate apriori information from data if not given and extend its range if not sufficient long enough
        if apriori is None:
            apriori = self.create_monthly_mean(data, sel_opts=sel_opts, sampling=sampling, time_dim=time_dim)
        apriori = self.extend_apriori(data, apriori, time_dim, sampling)

        # calculate FIR filter coefficients
        h = signal.firwin(order, cutoff_high, pass_zero="lowpass", fs=fs, window=window)
        length = len(h)

        # start loop on all timestamps
        dt = data.coords[time_dim].values
        res = xr.zeros_like(data)
        logging.info("start iteration")
        for i in range(0, len(dt)):
            t0 = dt[i]
            pd_date = pd.to_datetime(t0)
            if pd_date.day == 1 and pd_date.month == 1:
                print(t0)
            try:
                i_m = max(0, i - length)
                i_p = min(i + length, len(dt) - 2)
                t_hist = slice(dt[i_m], dt[i])
                t_fut = slice(dt[i + 1], dt[i_p + 1])
                tmp_hist = data.sel({time_dim: t_hist})
                tmp_fut = apriori.sel({time_dim: t_fut})
                tmp_comb = xr.concat([tmp_hist, tmp_fut], dim=time_dim)
                _padlen = int(min(padlen_factor, 1) * len(tmp_comb.coords[time_dim]))
                tmp_filter, _ = fir_filter(tmp_comb, fs, cutoff_high=cutoff_high, order=order, causal=False,
                                           padlen=_padlen, dim=var_dim, window=window, h=h)
                res.loc[{time_dim: t0}] = tmp_filter.loc[{time_dim: t0}]
                if i == 720 and self.plot_path is not None:
                    self.plot(data, tmp_comb, var_dim, time_dim, slice(dt[i_m], dt[i_p + 1]), t0, plot_index)
            except IndexError:
                res.loc[{time_dim: t0}] = np.nan
        return res, h, apriori

    @TimeTrackingWrapper
    def clim_filter_vectorized(self, data, fs, cutoff_high, order, apriori=None, padlen_factor=0.5, sel_opts=None,
                               sampling="1d", time_dim="datetime", var_dim="variables", window="hamming",
                               plot_index=None):

        # calculate apriori information from data if not given and extend its range if not sufficient long enough
        if apriori is None:
            apriori = self.create_monthly_mean(data, sel_opts=sel_opts, sampling=sampling, time_dim=time_dim)
        apriori = self.extend_apriori(data, apriori, time_dim, sampling)

        # calculate FIR filter coefficients
        h = signal.firwin(order, cutoff_high, pass_zero="lowpass", fs=fs, window=window)
        length = len(h)

        # create tmp dimension to apply filter, search for unused name
        new_dim = self._create_tmp_dimension(data)

        # combine historical data / observation [t0-length,t0] and climatological statistics [t0+1,t0+length]
        history = self._shift_data(data, range(int(-(length - 1) / 2), 1), time_dim, var_dim, new_dim)
        future = self._shift_data(apriori, range(1, int((length - 1) / 2) + 1), time_dim, var_dim, new_dim)
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        # filter_input_data = history.combine_first(future)
        # history.sel(datetime=slice("2010-11-01", "2011-04-01"),variables="o3").plot()
        # filter_input_data.sel(datetime=slice("2009-11-01", "2011-04-01"),variables="temp").plot()

        time_axis = filter_input_data.coords[time_dim]
        # apply vectorized fir filter along the tmp dimension
        kwargs = {"fs": fs, "cutoff_high": cutoff_high, "order": order,
                  "causal": False, "padlen": int(min(padlen_factor, 1) * length), "h": h}
        # with TimeTracking(name="numpy_vec"):
        #     filt = fir_filter_numpy_vectorized(filter_input_data, var_dim, new_dim, kwargs)
        # with TimeTracking(name="xr_apply_ufunc"):
        #     filt = xr.apply_ufunc(fir_filter_vectorized, filter_input_data, time_axis,
        #                           input_core_dims=[[new_dim], []], output_core_dims=[[new_dim]], vectorize=True,
        #                           kwargs=kwargs)
        with TimeTracking(name="convolve"):
            slicer = slice(int(-(length - 1) / 2), int((length - 1) / 2))
            filt = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_data.sel({new_dim: slicer}),
                                  input_core_dims=[[new_dim]],
                                  output_core_dims=[[new_dim]],
                                  vectorize=True,
                                  kwargs={"h": h})

        # plot
        if self.plot_path is not None:
            for i, time_pos in enumerate([0.25, 1.5, 2.75, 4]):  # [0.25, 1.5, 2.75, 4] x 365 days
                try:
                    pos = int(time_pos * 365 * fs)
                    filter_example = filter_input_data.isel({time_dim: pos})
                    t0 = filter_example.coords[time_dim].values
                    t_slice = filter_input_data.isel(
                        {time_dim: slice(pos - int((length - 1) / 2), pos + int((length - 1) / 2) + 1)}).coords[
                        time_dim].values
                    self.plot(data, filter_example, var_dim, time_dim, t_slice, t0, f"{plot_index}_{i}")
                except IndexError:
                    pass

        # select only values at tmp dimension 0 at each point in time
        res = filt.sel({new_dim: 0}, drop=True)
        # create result array with same shape like input data, gabs are filled by nans
        res_full = xr.ones_like(data) * np.nan
        res_full.loc[res.coords] = res
        return res_full, h, apriori

    def _tmp_analysis(self, data, apriori, var, var_dim, length, time_dim, new_dim, h):
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): sel data")
        d = data.sel({var_dim: [var]}).sel(datetime=slice("2007", "2010"))
        a = apriori.sel({var_dim: [var]}).sel(datetime=slice("2007", "2010"))

        # combine historical data / observation [t0-length,t0] and climatological statistics [t0+1,t0+length]
        history = self._shift_data(d, range(-length, 1), time_dim, var_dim, new_dim)

        future = self._shift_data(d, range(1, length), time_dim, var_dim, new_dim)
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
        with TimeTracking(name="convolve"):
            filt_nc = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_data,
                                     input_core_dims=[[new_dim]],
                                     output_core_dims=[[new_dim]],
                                     vectorize=True,
                                     kwargs={"h": h})

        future = self._shift_data(a, range(1, length), time_dim, var_dim, new_dim)
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
        with TimeTracking(name="convolve"):
            filt_t0 = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_data,
                                     input_core_dims=[[new_dim]],
                                     output_core_dims=[[new_dim]],
                                     vectorize=True,
                                     kwargs={"h": h})

        diff = (a - history.sel(window=slice(-24, 1)).mean(new_dim))
        future = self._shift_data(a, range(1, length), time_dim, var_dim, new_dim) - diff
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
        with TimeTracking(name="convolve"):
            filt_diff1d = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_data,
                                         input_core_dims=[[new_dim]],
                                         output_core_dims=[[new_dim]],
                                         vectorize=True,
                                         kwargs={"h": h})

        diff = (a - history.sel(window=slice(-24 * 7, 1)).mean(new_dim))
        future = self._shift_data(a, range(1, length), time_dim, var_dim, new_dim) - diff
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
        with TimeTracking(name="convolve"):
            filt_diff1w = xr.apply_ufunc(fir_filter_convolve_vectorized,
                                         filter_input_data,
                                         input_core_dims=[[new_dim]],
                                         output_core_dims=[[new_dim]],
                                         vectorize=True,
                                         kwargs={"h": h})

        diff = (a - history.sel(window=slice(-24 * 7, 1)).mean(new_dim))
        future = self._shift_data(a, range(1, length), time_dim, var_dim, new_dim)
        diff = xr.zeros_like(future) + diff
        lam = np.log(2) / (7 * 24)
        diff = diff * np.exp(- lam * diff.coords["window"])
        future = future - diff
        filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
        logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
        with TimeTracking(name="convolve"):
            filt_diff1w_decay = xr.apply_ufunc(fir_filter_convolve_vectorized,
                                               filter_input_data,
                                               input_core_dims=[[new_dim]],
                                               output_core_dims=[[new_dim]],
                                               vectorize=True,
                                               kwargs={"h": h})

        t0 = datetime.datetime.strptime("2009-07-15 00:00", "%Y-%m-%d %H:%M")
        delta = datetime.timedelta(hours=1)
        for i in range(int((length - 1) / 2)):
            plt.plot(-i, filt_nc.sel(datetime=t0 - i * delta, window=0), marker="+", color="black")
        filt_nc.sel(datetime=t0).plot(label="noncausal")
        filt_t0.sel(datetime=t0).plot(label="nodiff")
        filt_diff1d.sel(datetime=t0).plot(label="diff1d")
        filt_diff1w.sel(datetime=t0).plot(label="diff1w")
        filt_diff1w_decay.sel(datetime=t0).plot(label="diff1wdecay")
        plt.legend()

        for i in range(int((length - 1) / 2)):
            plt.plot(-i, filt_t0.sel(datetime=t0 - i * delta, window=0), marker="+", color="black")

        z = 1

    @TimeTrackingWrapper
    def clim_filter_vectorized_less_memory(self, data, fs, cutoff_high, order, apriori=None, padlen_factor=0.5,
                                           sel_opts=None,
                                           sampling="1d", time_dim="datetime", var_dim="variables", window="hamming",
                                           plot_index=None, minimum_length=None, new_dim="window", plot_dates=None):

        logging.info(f"{data.coords['Stations'].values[0]}: extend apriori")

        # calculate apriori information from data if not given and extend its range if not sufficient long enough
        if apriori is None:
            apriori = self.create_monthly_mean(data, sel_opts=sel_opts, sampling=sampling, time_dim=time_dim)
        apriori = apriori.astype(data.dtype)
        apriori = self.extend_apriori(data, apriori, time_dim, sampling)

        # calculate FIR filter coefficients
        h = signal.firwin(order, cutoff_high, pass_zero="lowpass", fs=fs, window=window)
        length = len(h)

        # create tmp dimension to apply filter, search for unused name
        # new_dim = self._create_tmp_dimension(data)

        # use filter length if no minimum is given, otherwise use minimum + half filter length for extension
        extend_length_history = length if minimum_length is None else minimum_length + int((length + 1) / 2)
        extend_length_future = int((length + 1) / 2) + 1

        # collect some data for visualization
        plot_pos = np.array([0.25, 1.5, 2.75, 4]) * 365 * fs
        if plot_dates is None:
            plot_dates = [data.isel({time_dim: int(pos)}).coords[time_dim].values for pos in plot_pos if
                          pos < len(data.coords[time_dim])]
        plot_data = []

        coll = []

        for var in reversed(data.coords[var_dim].values):
            # self._tmp_analysis(data, apriori, var, var_dim, length, time_dim, new_dim, h)
            logging.info(f"{data.coords['Stations'].values[0]} ({var}): sel data")

            _start = pd.to_datetime(data.coords[time_dim].min().values).year
            _end = pd.to_datetime(data.coords[time_dim].max().values).year
            filt_coll = []
            for _year in range(_start, _end + 1):
                logging.info(f"{data.coords['Stations'].values[0]} ({var}): year={_year}")

                time_slice = self._create_time_range_extend(_year, sampling, extend_length_history)
                d = data.sel({var_dim: [var], time_dim: time_slice})
                a = apriori.sel({var_dim: [var], time_dim: time_slice})
                if len(d.coords[time_dim]) == 0:  # no data at all for this year
                    continue

                # combine historical data / observation [t0-length,t0] and climatological statistics [t0+1,t0+length]
                # logging.info(f"{data.coords['Stations'].values[0]} ({var}): history")
                if new_dim not in d.coords:
                    history = self._shift_data(d, range(int(-extend_length_history), 1), time_dim, var_dim, new_dim)
                else:
                    history = d.sel({new_dim: slice(int(-extend_length_history), 0)})
                # logging.info(f"{data.coords['Stations'].values[0]} ({var}): future")
                # diff = (a - history.sel(window=slice(-24, 1)).mean(new_dim))
                if new_dim not in a.coords:
                    future = self._shift_data(a, range(1, extend_length_future), time_dim, var_dim, new_dim)
                    # future = self._shift_data(a, range(1, int((length - 1) / 2) + 1), time_dim, var_dim, new_dim) - diff
                else:
                    future = a.sel({new_dim: slice(1, extend_length_future)})
                # logging.info(f"{data.coords['Stations'].values[0]} ({var}): concat to filter input")

                filter_input_data = xr.concat([history.dropna(time_dim), future], dim=new_dim, join="left")
                try:
                    filter_input_data = filter_input_data.sel({time_dim: str(_year)})
                except KeyError:  # no valid data for this year
                    continue
                if len(filter_input_data.coords[time_dim]) == 0:  # no valid data for this year
                    continue

                logging.info(f"{data.coords['Stations'].values[0]} ({var}): start filter convolve")
                with TimeTracking(name="convolve"):
                    # slicer = slice(int(-(length - 1) / 2), int((length - 1) / 2))
                    filt = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_data,  # .sel({new_dim: slicer}),
                                          input_core_dims=[[new_dim]],
                                          output_core_dims=[[new_dim]],
                                          vectorize=True,
                                          kwargs={"h": h},
                                          output_dtypes=[d.dtype])

                if minimum_length is None:
                    filt_coll.append(filt.sel({new_dim: slice(-extend_length_history, 0)}, drop=True))
                else:
                    filt_coll.append(filt.sel({new_dim: slice(-minimum_length, 0)}, drop=True))

                # visualization
                for viz_date in set(plot_dates).intersection(filt.coords[time_dim].values):
                    try:
                        td_type = {"1d": "D", "1H": "h"}.get(sampling)
                        t_minus = viz_date + np.timedelta64(int(-extend_length_history), td_type)
                        t_plus = viz_date + np.timedelta64(int(extend_length_future), td_type)
                        if new_dim not in d.coords:
                            tmp_filter_data = self._shift_data(d.sel({time_dim: slice(t_minus, t_plus)}),
                                                               range(int(-extend_length_history),
                                                                     int(extend_length_future)),
                                                               time_dim, var_dim, new_dim).sel({time_dim: viz_date})
                        else:
                            # tmp_filter_data = d.sel({time_dim: viz_date,
                            #                          new_dim: slice(int(-extend_length_history), int(extend_length_future))})
                            tmp_filter_data = None
                        valid_range = range(int((length + 1) / 2) if minimum_length is None else minimum_length, 1)
                        plot_data.append({"t0": viz_date,
                                          "var": var,
                                          "filter_input": filter_input_data.sel({time_dim: viz_date}),
                                          "filter_input_nc": tmp_filter_data,
                                          "valid_range": valid_range,
                                          "time_range": d.sel(
                                              {time_dim: slice(t_minus, t_plus - np.timedelta64(1, td_type))}).coords[
                                              time_dim].values,
                                          "h": h,
                                          "new_dim": new_dim})
                    except:
                        pass

            # collect all filter results
            coll.append(xr.concat(filt_coll, time_dim))
            gc.collect()

            # plot
            # ToDo: enable plotting again
            # if self.plot_path is not None:
            #     for i, viz_data in enumerate(plot_data):
            #         self.plot_new(viz_data, data.sel({var_dim: [var]}), var_dim, time_dim, new_dim, f"{plot_index}_{i}",
            #                       sampling)

            # for i, time_pos in enumerate([0.25, 1.5, 2.75, 4]):  # [0.25, 1.5, 2.75, 4] x 365 days
            #     try:
            #
            #         plot_data = coll[-1]
            #         pos = int(time_pos * 365 * fs)
            #         filter_example = plot_data.isel({time_dim: pos})
            #         t0 = filter_example.coords[time_dim].values
            #
            #         slice_tmp = slice(pos - abs(plot_data.coords[new_dim].values.min()), pos + abs(plot_data.coords[new_dim].values.min()))
                #         t_slice = plot_data.isel({time_dim: slice_tmp}).coords[time_dim].values
                #         self.plot(data.sel({var_dim: [var]}), filter_example, var_dim, time_dim, t_slice, t0, f"{plot_index}_{i}")
                #     except IndexError:
                #         pass

        logging.info(f"{data.coords['Stations'].values[0]}: concat all variables")
        res = xr.concat(coll, var_dim)
        # create result array with same shape like input data, gabs are filled by nans
        logging.info(f"{data.coords['Stations'].values[0]}: create res_full")

        new_coords = {**{k: data.coords[k].values for k in data.coords if k != new_dim}, new_dim: res.coords[new_dim]}
        dims = [*data.dims, new_dim] if new_dim not in data.dims else data.dims
        res = res.transpose(*dims)
        # res_full = xr.DataArray(dims=dims, coords=new_coords)
        # res_full.loc[res.coords] = res
        # res_full.compute()
        res_full = res.broadcast_like(xr.DataArray(dims=dims, coords=new_coords))
        return res_full, h, apriori, plot_data

    @staticmethod
    def _create_time_range_extend(year, sampling, extend_length):
        td_type = {"1d": "D", "1H": "h"}.get(sampling)
        delta = np.timedelta64(extend_length + 1, td_type)
        start = np.datetime64(f"{year}-01-01") - delta
        end = np.datetime64(f"{year}-12-31") + delta
        return slice(start, end)

    @staticmethod
    def _create_tmp_dimension(data):
        new_dim = "window"
        count = 0
        while new_dim in data.dims:
            new_dim += new_dim
            count += 1
            if count > 10:
                raise ValueError("Could not create new dimension.")
        return new_dim

    def _shift_data(self, data, index_value, time_dim, squeeze_dim, new_dim):
        coll = []
        for i in index_value:
            coll.append(data.shift({time_dim: -i}))
        new_ind = self.create_index_array(new_dim, index_value, squeeze_dim)
        return xr.concat(coll, dim=new_ind)

    @staticmethod
    def create_index_array(index_name: str, index_value, squeeze_dim: str):
        ind = pd.DataFrame({'val': index_value}, index=index_value)
        res = xr.Dataset.from_dataframe(ind).to_array(squeeze_dim).rename({'index': index_name}).squeeze(
            dim=squeeze_dim,
            drop=True)
        res.name = index_name
        return res

    def _plot(self, sampling):
        new_dim = "window"
        h = None
        td_type = {"1d": "D", "1H": "h"}.get(sampling)
        if self.plot_path is None:
            return
        plot_folder = os.path.join(os.path.abspath(self.plot_path), "climFIR")
        if not os.path.exists(plot_folder):
            os.makedirs(plot_folder)

        rc_params = {'axes.labelsize': 'large',
                     'xtick.labelsize': 'large',
                     'ytick.labelsize': 'large',
                     'legend.fontsize': 'large',
                     'axes.titlesize': 'large',
                     }
        plt.rcParams.update(rc_params)

        plot_dict = {}
        for i, o in enumerate(range(len(self.plot_data))):
            plot_data = self.plot_data[i]
            for p_d in plot_data:
                var = p_d.get("var")
                t0 = p_d.get("t0")
                filter_input = p_d.get("filter_input")
                filter_input_nc = p_d.get("filter_input_nc")
                valid_range = p_d.get("valid_range")
                time_range = p_d.get("time_range")
                new_dim = p_d.get("new_dim")
                h = p_d.get("h")
                plot_dict_var = plot_dict.get(var, {})
                plot_dict_t0 = plot_dict_var.get(t0, {})
                plot_dict_order = {"filter_input": filter_input,
                                   "filter_input_nc": filter_input_nc,
                                   "valid_range": valid_range,
                                   "time_range": time_range,
                                   "order": o, "h": h}
                plot_dict_t0[i] = plot_dict_order
                plot_dict_var[t0] = plot_dict_t0
                plot_dict[var] = plot_dict_var

        for var, viz_date_dict in plot_dict.items():
            for it0, t0 in enumerate(viz_date_dict.keys()):
                viz_data = viz_date_dict[t0]
                residuum_true = None
                for ifilter in sorted(viz_data.keys()):
                    data = viz_data[ifilter]
                    filter_input = data["filter_input"]
                    filter_input_nc = data["filter_input_nc"] if residuum_true is None else residuum_true.sel(
                        {new_dim: filter_input.coords[new_dim]})
                    valid_range = data["valid_range"]
                    time_axis = data["time_range"]
                    # time_axis = pd.date_range(t_minus, t_plus, freq=sampling)
                    filter_order = data["order"]
                    h = data["h"]
                    t_minus = t0 + np.timedelta64(-int(1.5 * valid_range.start), td_type)
                    t_plus = t0 + np.timedelta64(int(0.5 * valid_range.start), td_type)
                    fig, ax = plt.subplots()
                    ax.axvspan(t0 + np.timedelta64(-valid_range.start, td_type),
                               t0 + np.timedelta64(valid_range.stop - 1, td_type), color="whitesmoke",
                               label="valid area")
                    ax.axvline(t0, color="lightgrey", lw=6, label="time of interest ($t_0$)")

                    # original data
                    ax.plot(time_axis, filter_input_nc.values.flatten(), color="darkgrey", linestyle="dashed",
                            label="original")

                    # clim apriori
                    if ifilter == 0:
                        d_tmp = filter_input.sel(
                            {new_dim: slice(0, filter_input.coords[new_dim].values.max())}).values.flatten()
                    else:
                        d_tmp = filter_input.values.flatten()
                    ax.plot(time_axis[len(time_axis) - len(d_tmp):], d_tmp, color="darkgrey", linestyle="solid",
                            label="estimated future")

                    # clim filter response
                    filt = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input,
                                          input_core_dims=[[new_dim]],
                                          output_core_dims=[[new_dim]],
                                          vectorize=True,
                                          kwargs={"h": h},
                                          output_dtypes=[filter_input.dtype])
                    ax.plot(time_axis, filt.values.flatten(), color="black", linestyle="solid",
                            label="clim filter response", linewidth=2)
                    residuum_estimated = filter_input - filt

                    # ideal filter response
                    filt = xr.apply_ufunc(fir_filter_convolve_vectorized, filter_input_nc,
                                          input_core_dims=[[new_dim]],
                                          output_core_dims=[[new_dim]],
                                          vectorize=True,
                                          kwargs={"h": h},
                                          output_dtypes=[filter_input.dtype])
                    ax.plot(time_axis, filt.values.flatten(), color="black", linestyle="dashed",
                            label="ideal filter response", linewidth=2)
                    residuum_true = filter_input_nc - filt

                    # set title, legend, and save plot
                    ax_start = max(t_minus, time_axis[0])
                    ax_end = min(t_plus, time_axis[-1])
                    ax.set_xlim((ax_start, ax_end))
                    plt.title(f"Input of ClimFilter ({str(var)})")
                    plt.legend()
                    fig.autofmt_xdate()
                    plt.tight_layout()
                    plot_name = os.path.join(plot_folder,
                                             f"climFIR_{self.plot_name}_{str(var)}_{it0}_{ifilter}.pdf")
                    plt.savefig(plot_name, dpi=300)
                    plt.close('all')

                    # plot residuum
                    fig, ax = plt.subplots()
                    ax.axvspan(t0 + np.timedelta64(-valid_range.start, td_type),
                               t0 + np.timedelta64(valid_range.stop - 1, td_type), color="whitesmoke",
                               label="valid area")
                    ax.axvline(t0, color="lightgrey", lw=6, label="time of interest ($t_0$)")
                    ax.plot(time_axis, residuum_true.values.flatten(), color="black", linestyle="dashed",
                            label="ideal filter residuum", linewidth=2)
                    ax.plot(time_axis, residuum_estimated.values.flatten(), color="black", linestyle="solid",
                            label="clim filter residuum", linewidth=2)
                    ax.set_xlim((ax_start, ax_end))
                    plt.title(f"Residuum of ClimFilter ({str(var)})")
                    plt.legend()
                    fig.autofmt_xdate()
                    plt.tight_layout()
                    plot_name = os.path.join(plot_folder,
                                             f"climFIR_{self.plot_name}_{str(var)}_{it0}_{ifilter}_residuum.pdf")
                    plt.savefig(plot_name, dpi=300)
                    plt.close('all')

    def plot_new(self, viz_data, orig_data, var_dim, time_dim, new_dim, plot_index, sampling):
        try:
            td_type = {"1d": "D", "1H": "h"}.get(sampling)
            filter_example = viz_data["filt"]
            filter_input = viz_data["filter_input"]
            filter_nc = viz_data["filt_nc"]
            valid_range = viz_data["valid_range"]
            t0 = viz_data["t0"]
            t_minus = t0 + np.timedelta64(filter_input.coords[new_dim].values.min(), td_type)
            t_plus = t0 + np.timedelta64(filter_input.coords[new_dim].values.max(), td_type)
            t_slice = slice(t_minus, t_plus)
            data = orig_data.sel({time_dim: t_slice})
            plot_folder = os.path.join(os.path.abspath(self.plot_path), "climFIR")
            if not os.path.exists(plot_folder):
                os.makedirs(plot_folder)

            for var in data.coords[var_dim]:
                time_axis = data.sel({var_dim: var, time_dim: t_slice}).coords[time_dim].values
                rc_params = {'axes.labelsize': 'large',
                             'xtick.labelsize': 'large',
                             'ytick.labelsize': 'large',
                             'legend.fontsize': 'large',
                             'axes.titlesize': 'large',
                             }
                plt.rcParams.update(rc_params)
                fig, ax = plt.subplots()

                ax.axvspan(t0 + np.timedelta64(-valid_range.start, td_type),
                           t0 + np.timedelta64(valid_range.stop - 1, td_type), color="whitesmoke", label="valid area")

                ax.axvline(t0, color="lightgrey", lw=6, label="time of interest ($t_0$)")
                ax.plot(time_axis, data.sel({var_dim: var, time_dim: t_slice}).values.flatten(),
                        color="darkgrey", linestyle="dashed", label="original")
                d_tmp = filter_input.sel(
                    {var_dim: var, new_dim: slice(0, filter_input.coords[new_dim].values.max())}).values.flatten()
                # ax.plot(time_axis[len(time_axis) - len(d_tmp):], d_tmp, color="darkgrey", linestyle=(0 ,(1, 1)), label="filter input")
                ax.plot(time_axis[len(time_axis) - len(d_tmp):], d_tmp, color="darkgrey", linestyle="solid",
                        label="estimated future")
                # data.sel({var_dim: var, time_dim: time_dim_slice}).plot()
                # tmp_comb.sel({var_dim: var}).plot()
                # d_filt = filter_example.sel({var_dim: var}).values.flatten()
                ax.plot(time_axis, filter_example.sel({var_dim: var}).values.flatten(),
                        color="black", linestyle="solid", label="filter response", linewidth=2)
                ax.plot(time_axis, filter_nc.sel({var_dim: var}).values.flatten(),
                        color="black", linestyle="dashed", label="ideal filter response", linewidth=2)
                plt.title(f"Input of ClimFilter ({str(var.values)})")
                plt.legend()
                fig.autofmt_xdate()
                plt.tight_layout()
                plot_name = os.path.join(plot_folder, f"climFIR_{self.plot_name}_{str(var.values)}_{plot_index}.pdf")
                plt.savefig(plot_name, dpi=300)
                plt.close('all')
        except:
            pass

    def plot(self, data, tmp_comb, var_dim, time_dim, time_dim_slice, t0, plot_index):
        try:
            plot_folder = os.path.join(os.path.abspath(self.plot_path), "climFIR")
            if not os.path.exists(plot_folder):
                os.makedirs(plot_folder)
            for var in data.coords[var_dim]:
                time_axis = data.sel({var_dim: var, time_dim: time_dim_slice}).coords[time_dim].values
                rc_params = {'axes.labelsize': 'large',
                             'xtick.labelsize': 'large',
                             'ytick.labelsize': 'large',
                             'legend.fontsize': 'large',
                             'axes.titlesize': 'large',
                             }
                plt.rcParams.update(rc_params)
                fig, ax = plt.subplots()
                ax.axvline(t0, color="lightgrey", lw=6, label="time of interest ($t_0$)")
                ax.plot(time_axis, data.sel({var_dim: var, time_dim: time_dim_slice}).values.flatten(),
                        color="darkgrey", linestyle="--", label="original")
                d_filt = tmp_comb.sel({var_dim: var}).values.flatten()
                ax.plot(time_axis[:len(d_filt)], d_filt, color="black", label="filter input")
                # data.sel({var_dim: var, time_dim: time_dim_slice}).plot()
                # tmp_comb.sel({var_dim: var}).plot()
                plt.title(f"Input of ClimFilter ({str(var.values)})")
                plt.legend()
                fig.autofmt_xdate()
                plt.tight_layout()
                plot_name = os.path.join(plot_folder, f"climFIR_{self.plot_name}_{str(var.values)}_{plot_index}.pdf")
                plt.savefig(plot_name, dpi=300)
                plt.close('all')
        except:
            pass

    @property
    def filter_coefficients(self):
        return self._h

    @property
    def filtered_data(self):
        return self._filtered

    @property
    def apriori_data(self):
        return self._apriori

    @property
    def initial_apriori_data(self):
        return self.apriori_data[0]


def fir_filter(data, fs, order=5, cutoff_low=None, cutoff_high=None, window="hamming", dim="variables", h=None,
               causal=True, padlen=None):
    """Expects xarray."""
    if h is None:
        cutoff = []
        if cutoff_low is not None:
            cutoff += [cutoff_low]
        if cutoff_high is not None:
            cutoff += [cutoff_high]
        if len(cutoff) == 2:
            filter_type = "bandpass"
        elif len(cutoff) == 1 and cutoff_low is not None:
            filter_type = "highpass"
        elif len(cutoff) == 1 and cutoff_high is not None:
            filter_type = "lowpass"
        else:
            raise ValueError("Please provide either cutoff_low or cutoff_high.")
        h = signal.firwin(order, cutoff, pass_zero=filter_type, fs=fs, window=window)
    filtered = xr.ones_like(data)
    for var in data.coords[dim]:
        d = data.sel({dim: var}).values.flatten()
        if causal:
            y = signal.lfilter(h, 1., d)
        else:
            padlen = padlen if padlen is not None else 3 * len(h)
            y = signal.filtfilt(h, 1., d, padlen=padlen)
        filtered.loc[{dim: var}] = y
    return filtered, h


def fir_filter_numpy_vectorized(filter_input_data, var_dim, new_dim, kwargs):
    filt_np = xr.DataArray(np.nan, coords=filter_input_data.coords)
    for var in filter_input_data.coords[var_dim]:
        logging.info(
            f"{filter_input_data.coords['Stations'].values[0]}: {str(var.values)}")  # ToDo must be removed, just for debug
        a = np.apply_along_axis(fir_filter_vectorized, filter_input_data.dims.index(new_dim),
                                filter_input_data.sel({var_dim: var}).values, **kwargs)
        filt_np.loc[{var_dim: var}] = a
    return filt_np


def fir_filter_convolve_vectorized(data, h):
    return signal.convolve(data, h, mode='same', method="direct") / sum(h)


def fir_filter_vectorized(data, time_stamp=None, fs=1, order=5, cutoff_low=None, cutoff_high=None, window="hamming",
                          h=None,
                          causal=True,
                          padlen=None):
    """Expects numpy array."""
    if time_stamp is not None:
        pd_date = pd.to_datetime(time_stamp)
        if pd_date.day == 1 and pd_date.month in [1, 7]:
            logging.info(time_stamp)
    # sel = ~np.isnan(data)
    # res = np.empty_like(data)
    if h is None:
        cutoff = []
        if cutoff_low is not None:
            cutoff += [cutoff_low]
        if cutoff_high is not None:
            cutoff += [cutoff_high]
        if len(cutoff) == 2:
            filter_type = "bandpass"
        elif len(cutoff) == 1 and cutoff_low is not None:
            filter_type = "highpass"
        elif len(cutoff) == 1 and cutoff_high is not None:
            filter_type = "lowpass"
        else:
            raise ValueError("Please provide either cutoff_low or cutoff_high.")
        h = signal.firwin(order, cutoff, pass_zero=filter_type, fs=fs, window=window)
    if causal:
        # y = signal.lfilter(h, 1., data[sel])
        y = signal.lfilter(h, 1., data)
    else:
        padlen = padlen if padlen is not None else 3 * len(h)
        # print(sum(sel))
        # if sum(sel) <= padlen:
        #     y = np.empty_like(data[sel])
        # else:
        #     y = signal.filtfilt(h, 1., data[sel], padlen=padlen)
        y = signal.filtfilt(h, 1., data, padlen=padlen)
    # res[sel] = y
    # return res
    return y


def fir_filter_vectorized_short(data, h=None, padlen=None):
    """Expects numpy array."""
    y = signal.filtfilt(h, 1., data, padlen=padlen)
    return y


class KolmogorovZurbenkoBaseClass:

    def __init__(self, df, wl, itr, is_child=False, filter_dim="window"):
        """
        It create the variables associate with the Kolmogorov-Zurbenko-filter.

        Args:
            df(pd.DataFrame, None): time series of a variable
            wl(list of int): window length
            itr(list of int): number of iteration
        """
        self.df = df
        self.filter_dim = filter_dim
        self.wl = to_list(wl)
        self.itr = to_list(itr)
        if abs(len(self.wl) - len(self.itr)) > 0:
            raise ValueError("Length of lists for wl and itr must agree!")
        self._isChild = is_child
        self.child = self.set_child()
        self.type = type(self).__name__

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoBaseClass(None, self.wl[1:], self.itr[1:], True, self.filter_dim)
        else:
            return None

    def kz_filter(self, df, m, k):
        pass

    def spectral_calc(self):
        df_start = self.df
        kz = self.kz_filter(df_start, self.wl[0], self.itr[0])
        filtered = self.subtract(df_start, kz)
        # case I: no child avail -> return kz and remaining
        if self.child is None:
            return [kz, filtered]
        # case II: has child -> return current kz and all child results
        else:
            self.child.df = filtered
            kz_next = self.child.spectral_calc()
            return [kz] + kz_next

    @staticmethod
    def subtract(minuend, subtrahend):
        try:  # pandas implementation
            return minuend.sub(subtrahend, axis=0)
        except AttributeError:  # general implementation
            return minuend - subtrahend

    def run(self):
        return self.spectral_calc()

    def transfer_function(self):
        m = self.wl[0]
        k = self.itr[0]
        omega = np.linspace(0.00001, 0.15, 5000)
        return omega, (np.sin(m * np.pi * omega) / (m * np.sin(np.pi * omega))) ** (2 * k)

    def omega_null(self, alpha=0.5):
        a = np.sqrt(6) / np.pi
        b = 1 / (2 * np.array(self.itr))
        c = 1 - alpha ** b
        d = np.array(self.wl) ** 2 - alpha ** b
        return a * np.sqrt(c / d)

    def period_null(self, alpha=0.5):
        return 1. / self.omega_null(alpha)

    def period_null_days(self, alpha=0.5):
        return self.period_null(alpha) / 24.

    def plot_transfer_function(self, fig=None, name=None):
        if fig is None:
            fig = plt.figure()
        omega, transfer_function = self.transfer_function()
        if self.child is not None:
            transfer_function_child = self.child.plot_transfer_function(fig)
        else:
            transfer_function_child = transfer_function * 0
        plt.semilogx(omega, transfer_function - transfer_function_child,
                     label="m={:3.0f}, k={:3.0f}, T={:6.2f}d".format(self.wl[0],
                                                                     self.itr[0],
                                                                     self.period_null_days()))
        plt.axvline(x=self.omega_null())
        if not self._isChild:
            locs, labels = plt.xticks()
            plt.xticks(locs, np.round(1. / (locs * 24), 1))
            plt.xlim([0.00001, 0.15])
            plt.legend()
            if name is None:
                plt.show()
            else:
                plt.savefig(name)
        else:
            return transfer_function


class KolmogorovZurbenkoFilterMovingWindow(KolmogorovZurbenkoBaseClass):

    def __init__(self, df, wl: Union[list, int], itr: Union[list, int], is_child=False, filter_dim="window",
                 method="mean", percentile=0.5):
        """
        It create the variables associate with the KolmogorovZurbenkoFilterMovingWindow class.

        Args:
            df(pd.DataFrame, xr.DataArray): time series of a variable
            wl: window length
            itr: number of iteration
        """
        self.valid_methods = ["mean", "percentile", "median", "max", "min"]
        if method not in self.valid_methods:
            raise ValueError("Method '{}' is not supported. Please select from [{}].".format(
                method, ", ".join(self.valid_methods)))
        else:
            self.method = method
            if percentile > 1 or percentile < 0:
                raise ValueError("Percentile must be in range [0, 1]. Given was {}!".format(percentile))
            else:
                self.percentile = percentile
        super().__init__(df, wl, itr, is_child, filter_dim)

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoFilterMovingWindow(self.df, self.wl[1:], self.itr[1:], is_child=True,
                                                        filter_dim=self.filter_dim, method=self.method,
                                                        percentile=self.percentile)
        else:
            return None

    @TimeTrackingWrapper
    def kz_filter_new(self, df, wl, itr):
        """
        It passes the low frequency time series.

        If filter method is from mean, max, min this method will call construct and rechunk before the actual
        calculation to improve performance. If filter method is either median or percentile this approach is not
        applicable and depending on the data and window size, this method can become slow.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            for i in np.arange(0, itr):
                print(i)
                rolling = df_itr.chunk().rolling(**kwargs)
                if self.method not in ["percentile", "median"]:
                    rolling = rolling.construct("construct").chunk("auto")
                if self.method == "median":
                    df_mv_avg_tmp = rolling.median()
                elif self.method == "percentile":
                    df_mv_avg_tmp = rolling.quantile(self.percentile)
                elif self.method == "max":
                    df_mv_avg_tmp = rolling.max("construct")
                elif self.method == "min":
                    df_mv_avg_tmp = rolling.min("construct")
                else:
                    df_mv_avg_tmp = rolling.mean("construct")
                df_itr = df_mv_avg_tmp.compute()
                del df_mv_avg_tmp, rolling
                gc.collect()
            return df_itr
        except ValueError:
            raise ValueError

    @TimeTrackingWrapper
    def kz_filter(self, df, wl, itr):
        """
        It passes the low frequency time series.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        import warnings
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            iter_vars = df_itr.coords["variables"].values
            for var in iter_vars:
                df_itr_var = df_itr.sel(variables=[var])
                for _ in np.arange(0, itr):
                    df_itr_var = df_itr_var.chunk()
                    rolling = df_itr_var.rolling(**kwargs)
                    if self.method == "median":
                        df_mv_avg_tmp = rolling.median()
                    elif self.method == "percentile":
                        df_mv_avg_tmp = rolling.quantile(self.percentile)
                    elif self.method == "max":
                        df_mv_avg_tmp = rolling.max()
                    elif self.method == "min":
                        df_mv_avg_tmp = rolling.min()
                    else:
                        df_mv_avg_tmp = rolling.mean()
                    df_itr_var = df_mv_avg_tmp.compute()
                df_itr.loc[{"variables": [var]}] = df_itr_var
            return df_itr
        except ValueError:
            raise ValueError
